// Program gCodeOptimise
// Include Libraries
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <conio.h>
#include <stdbool.h>
// Uses Daniil Guitelson's BGI library
#include "graphics.h" // -lBGI -lgdi32

typedef struct Block {
   int G,F,S,T,M;
   double X,Y,Z;
   int tag;
   int seg;
} Block;

int main(int argc,char **argv) {
  printf("gCode Optimiser - Written by Alan Cooper (agp.cooper@gmail.com)\n");

  bool laserFlag=false;
  bool fileInFlag=false;
  bool fileOutFlag=false;
  bool g0Flag=false;
  bool g1Flag=false;
  char fileInName[33]="";
  char fileOutName[33]="";
  int blocks=0;
  double tolerance=0.01;
  int feed0=60;
  int feed1=60;
  double length=0.0;
  double area=0.0;
  struct Block** block=NULL;
  int i,j,k,count;
  bool flag;

  // Get parameters
  {
    int k;
    double f1;
    for (k=1;k<argc;k++) {
      if (strcmp(argv[k],"-i")==0) {
        fileInFlag=true;
        if (k+1<argc) strcpy(fileInName,argv[k+1]);
      }
      if (strcmp(argv[k],"-I")==0) {
        fileInFlag=true;
        if (k+1<argc) strcpy(fileInName,argv[k+1]);
      }
      if (strcmp(argv[k],"-o")==0) {
        fileOutFlag=true;
        if (k+1<argc) strcpy(fileOutName,argv[k+1]);
      }
      if (strcmp(argv[k],"-O")==0) {
        fileOutFlag=true;
        if (k+1<argc) strcpy(fileOutName,argv[k+1]);
      }
      if (strcmp(argv[k],"-t")==0) {
        if (k+1<argc) sscanf(argv[k+1],"%lf",&f1);
        tolerance=f1;
        if (tolerance<=0) tolerance=0.001;
      }
      if (strcmp(argv[k],"-T")==0) {
        if (k+1<argc) sscanf(argv[k+1],"%lf",&f1);
        tolerance=f1;
        if (tolerance<=0) tolerance=0.001;
      }
      if (strcmp(argv[k],"-l")==0) {
        laserFlag=true;
      }
      if (strcmp(argv[k],"-L")==0) {
        laserFlag=true;
      }
      if (strcmp(argv[k],"-g0")==0) {
        if (k+1<argc) sscanf(argv[k+1],"%lf",&f1);
        feed0=(int)f1;
        if (feed0<=0) feed0=60;
        g0Flag=true;
      }
      if (strcmp(argv[k],"-G0")==0) {
        if (k+1<argc) sscanf(argv[k+1],"%lf",&f1);
        feed0=(int)f1;
        if (feed0<=0) feed0=60;
        g0Flag=true;
      }
      if (strcmp(argv[k],"-g1")==0) {
        if (k+1<argc) sscanf(argv[k+1],"%lf",&f1);
        feed1=(int)f1;
        if (feed1<=0) feed1=60;
        g1Flag=true;
      }
      if (strcmp(argv[k],"-G1")==0) {
        if (k+1<argc) sscanf(argv[k+1],"%lf",&f1);
        feed1=(int)f1;
        if (feed1<=0) feed1=60;
        g1Flag=true;
      }
    }
  }

  if ((!fileInFlag)||(!fileOutFlag)) {
    printf("Usage -i \"Input File\" -o \"Output File\" -t \"tolerance\"");
    printf(" -l \"laser\" -g0 \"overide G0 feed rate\" -g1 \"overide G1 feed rate\"\n");
    printf("Press any key to exit");
    getchar();
  } else {
    // Load the block model
    char Line[128];
    char *tmp;
    double f1;

    // Import the blocks
    FILE *fin=fopen(fileInName,"r");
    if (fin==NULL) {
      printf("Unable to read input file\n");
      return(1);
    }
    // Read the number of blocks or records
    blocks=0;
    while (fgets(Line,sizeof(Line),fin)!=NULL) blocks++;
    // Build the storage for the blocks
    block=(struct Block**)malloc((blocks+1)*sizeof(struct Block*));
    for (i=0;i<=blocks;i++) block[i]=(struct Block*)malloc(sizeof(struct Block));

    // Load the blocks
    rewind(fin);
    i=0;
    block[i]->G=0;
    block[i]->X=0.0;
    block[i]->Y=0.0;
    block[i]->Z=0.0;
    block[i]->F=-1;
    block[i]->S=-1;
    block[i]->T=-1;
    block[i]->M=-1;
    block[i]->tag=0;
    while (fgets(Line,sizeof(Line),fin)!=NULL) {
      i++;
      block[i]->G=block[i-1]->G;
      block[i]->X=block[i-1]->X;
      block[i]->Y=block[i-1]->Y;
      block[i]->Z=block[i-1]->Z;
      block[i]->F=-1;
      block[i]->S=-1;
      block[i]->T=-1;
      block[i]->M=-1;
      block[i]->tag=0;
      for (j=0;Line[j];j++) Line[j]=toupper(Line[j]);
      tmp=strchr(Line,'G');
      if (tmp!=NULL) {
        sscanf((tmp+1),"%lf",&f1);
        block[i]->G=(int)f1;
      }
      tmp=strchr(Line,'X');
      if (tmp!=NULL) {
        sscanf((tmp+1),"%lf",&f1);
        block[i]->X=f1;
      }
      tmp=strchr(Line,'Y');
      if (tmp!=NULL) {
        sscanf((tmp+1),"%lf",&f1);
        block[i]->Y=f1;
      }
      tmp=strchr(Line,'Z');
      if (tmp!=NULL) {
        sscanf((tmp+1),"%lf",&f1);
        block[i]->Z=f1;
      }
      tmp=strchr(Line,'F');
      if (tmp!=NULL) {
        sscanf((tmp+1),"%lf",&f1);
        if (f1<0) f1=0;
        block[i]->F=(int)(f1);
      }
      tmp=strchr(Line,'S');
      if (tmp!=NULL) {
        sscanf((tmp+1),"%lf",&f1);
        if (f1<0) f1=0;
        block[i]->S=(int)(f1);
      }
      tmp=strchr(Line,'T');
      if (tmp!=NULL) {
        sscanf((tmp+1),"%lf",&f1);
        if (f1<0) f1=0;
        block[i]->T=(int)(f1);
      }
      tmp=strchr(Line,'M');
      if (tmp!=NULL) {
        sscanf((tmp+1),"%lf",&f1);
        if (f1<0) f1=0;
        block[i]->M=(int)(f1);
      }
    }
    fclose(fin);
    printf("Blocks imported %d\n",blocks);

    // Filter
    for (i=0;i<=blocks;i++) block[i]->tag=0;
    for (k=1;k<=10;k++) {
      count=0;
      for (i=0;i<=blocks;i++) {
        count++;
        if (block[i]->G!=1) count=0;
        if (count==3) {
          if ((fabs(block[i-2]->Z-block[i-1]->Z)<tolerance)&&(fabs(block[i-1]->Z-block[i]->Z)<tolerance)) {
            area=fabs((block[i-1]->Y-block[i-2]->Y)*(block[i]->X-block[i-2]->X)-(block[i]->Y-block[i-2]->Y)*(block[i-1]->X-block[i-2]->X));
            length=sqrt((block[i]->Y-block[i-2]->Y)*(block[i]->Y-block[i-2]->Y)+(block[i]->X-block[i-2]->X)*(block[i]->X-block[i-2]->X));
            if (length*tolerance>area) block[i-1]->tag=1;
            if (tolerance>length) block[i-1]->tag=1;
          }
          if (block[i-2]->F!=block[i-1]->F) block[i-1]->tag=0;
          if (block[i-2]->S!=block[i-1]->S) block[i-1]->tag=0;
          if (block[i-2]->T!=block[i-1]->T) block[i-1]->tag=0;
          if (block[i-2]->M!=block[i-1]->M) block[i-1]->tag=0;
          count=1;
        }
      }
      j=0;
      for (i=0;i<=blocks;i++) {
        if (block[i]->tag==0) {
          block[j]->G=block[i]->G;
          block[j]->X=block[i]->X;
          block[j]->Y=block[i]->Y;
          block[j]->Z=block[i]->Z;
          block[j]->F=block[i]->F;
          block[j]->S=block[i]->S;
          block[j]->T=block[i]->T;
          block[j]->M=block[i]->M;
          block[j]->seg=block[i]->seg;
          block[j]->tag=0;
          j++;
        }
      }
      blocks=j-1;
    }
    printf("Blocks exported %d\n",blocks);

    // Export the blocks
    FILE *fout=fopen(fileOutName,"w");
    if (fout==NULL) {
      printf("Unable to write output file\n");
      return(1);
    }

    if (g0Flag) {
      fprintf(fout,"G0F%d\n",feed0);
    }
    if (g1Flag) {
      fprintf(fout,"G1F%d\n",feed1);
    }
    for (i=1;i<=blocks;i++) {
      // Export the data read
      flag=false;
      if (block[i]->T>=0) {
        fprintf(fout,"T%d",block[i]->T);
        flag=true;
      }
      if (!laserFlag) {
        if (block[i]->M==3) {
          fprintf(fout,"M%d",block[i]->M);
          flag=true;
        }
      } else {
        if ((block[i-1]->G!=1)&&(block[i]->G==1)) {
          fprintf(fout,"M%d",3);
          flag=true;
        }
      }
      if (block[i]->M>=0) {
        if ((block[i]->M!=0)&&(block[i]->M!=1)&&(block[i]->M!=2)&(block[i]->M!=3)) {
          if ((block[i]->M!=5)&&(block[i]->M!=9)&&(block[i]->M!=30)&&(block[i]->M!=60)) {
            fprintf(fout,"M%d",block[i]->M);
            flag=true;
          }
        }
      }
      if ((block[i]->G==0)||(block[i]->G==1)) {
        if (block[i]->G!=block[i-1]->G) {
          fprintf(fout,"G%d",block[i]->G);
          flag=true;
        }
        if (block[i]->X!=block[i-1]->X) {
          fprintf(fout,"X%0.3lf",block[i]->X);
          flag=true;
        }
        if (block[i]->Y!=block[i-1]->Y) {
          fprintf(fout,"Y%0.3lf",block[i]->Y);
          flag=true;
        }
        if ((!laserFlag)&&(block[i]->Z!=block[i-1]->Z)) {
          fprintf(fout,"Z%0.3lf",block[i]->Z);
          flag=true;
        }
      } else {
        fprintf(fout,"G%d",block[i]->G);
        flag=true;
      }
      if (block[i]->F>=0) {
        if ((block[i]->G==1)&&(!g1Flag)) {
          fprintf(fout,"F%d",block[i]->F);
          flag=true;
        }
        if ((block[i]->G==0)&&(!g0Flag)) {
          fprintf(fout,"F%d",block[i]->F);
          flag=true;
        }
      }
      if (block[i]->S>=0) {
        fprintf(fout,"S%d",block[i]->S);
        flag=true;
      }
      if ((block[i]->M==0)||(block[i]->M==1)||(block[i]->M==2)||(block[i]->M==30)) {
        if (flag) fprintf(fout,"\n");
        fprintf(fout,"M%d",block[i]->M);
        flag=true;
      }
      if ((block[i]->M==9)||(block[i]->M==60)) {
        if (flag) fprintf(fout,"\n");
        fprintf(fout,"M%d",block[i]->M);
        flag=true;
      }
      if ((!laserFlag)&&(block[i]->M==5)) {
        if (flag) fprintf(fout,"\n");
        fprintf(fout,"M%d",block[i]->M);
        flag=true;
      }
      if (i<blocks) {
        if ((laserFlag)&&(block[i]->G==1)&&(block[i+1]->G!=1)) {
          if (flag) fprintf(fout,"\n");
          fprintf(fout,"M%d",5);
          flag=true;
        }
      }

      if (flag) fprintf(fout,"\n");
    }
    fclose(fout);

    {
      int gd=CUSTOM, gm=CUSTOM_MODE(700,700);
      initgraph(&gd, &gm, "");

      int X1,Y1,X2,Y2;
      double scale,xmin,ymin,xmax,ymax;

      // Find the maximum and minimum data range
      xmin=block[0]->X;
      ymin=block[0]->Y;
      xmax=block[0]->X;
      ymax=block[0]->Y;
      for (i=1;i<=blocks;i++) {
        if (xmin>block[i]->X) xmin=block[i]->X;
        if (xmax<block[i]->X) xmax=block[i]->X;
        if (ymin>block[i]->Y) ymin=block[i]->Y;
        if (ymax<block[i]->Y) ymax=block[i]->Y;
      }
      scale=1.1*(xmax-xmin>ymax-ymin?xmax-xmin:ymax-ymin);

      // Display all edges
      setcolor(WHITE);
      for (i=0;i<blocks;i++) {
        if ((block[i]->G==1)&&(block[i+1]->G==1)) {
          X1=(int)((block[i]->X-(xmin+xmax)/2)*700/scale+350);
          Y1=(int)((-block[i]->Y+(ymin+ymax)/2)*700/scale+350);
          X2=(int)((block[i+1]->X-(xmin+xmax)/2)*700/scale+350);
          Y2=(int)((-block[i+1]->Y+(ymin+ymax)/2)*700/scale+350);
          line(X1,Y1,X2,Y2);
        }
      }
      // Display all edges
      setcolor(RED);
      for (i=0;i<blocks;i++) {
        if ((block[i]->G==0)&&(block[i+1]->G==0)) {
          X1=(int)((block[i]->X-(xmin+xmax)/2)*700/scale+350);
          Y1=(int)((-block[i]->Y+(ymin+ymax)/2)*700/scale+350);
          X2=(int)((block[i+1]->X-(xmin+xmax)/2)*700/scale+350);
          Y2=(int)((-block[i+1]->Y+(ymin+ymax)/2)*700/scale+350);
          line(X1,Y1,X2,Y2);
        }
      }

      getchar();
      closegraph();
    }
  }
  return(0);
}
